unit WorkerForm;

interface

uses
  Windows, Messages, SysUtils, Classes, Graphics, Controls, Forms, Dialogs,
  ComCtrls, DelimitedStringList, TUtil32, BDE, Db, DBTables, StdCtrls;

const
    tm_Init = 0;
    tm_NextFile = 1;

    rt_Verify = 1;
    rt_Rebuild = 2;
    rt_Auto = 3;

type
    TBDEUtil = class;

  TfWorkerForm = class(TForm)
    pbVerifyHeader: TProgressBar;
    pbVerifyIndex: TProgressBar;
    pbVerifyData: TProgressBar;
    pbRebuild: TProgressBar;
    Table1: TTable;
    Table1testfield: TIntegerField;
    Label1: TLabel;
    Label2: TLabel;
    Label3: TLabel;
    Label4: TLabel;
    mNotes: TMemo;
    tOpenAttempt: TTable;
    procedure FormCreate(Sender: TObject);
    procedure FormDestroy(Sender: TObject);
    procedure FormClose(Sender: TObject; var Action: TCloseAction);
  private
    { Private declarations }
    fClosing: Boolean;
    BDEUtil: TBDEUtil;
    TablesList: TDelimitedStringList;
    CurrentFileIndex: integer;
    successcount, failedcount: integer;
    fCloseWhenDone: boolean;
    fDefaultRepairType: integer;
    procedure WMTimer( var Msg : TMessage ); message wm_timer;
    procedure AddLogMessage(S: string);
    procedure ResetProgressBar(which: integer);
  public
    { Public declarations }
    iFld, iIdx, iSec, iVal, iRI, iOptP, iOptD: word;

    property Closing: Boolean read fClosing write fClosing default False;
    property CloseWhenDone: boolean read fCloseWhenDone write fCloseWhenDone;
    property DefaultRepairType: integer read fDefaultRepairType write fDefaultRepairType;
    function VerifyTable(tablename: string; var TableMessage: integer): boolean;
    function RebuildTable(tablename: string; var TableMessage: integer): boolean;
  end;

  TBDEUtil = class
    CbInfo: TUVerifyCallback;
    TUProps: CURProps;
    hDb: hDBIDb;
    vhTSes: hTUSes;
    constructor Create;
    destructor Destroy; override;
    function GetTCursorProps(szTable: String): Boolean;
    procedure RegisterCallBack;
    procedure UnRegisterCallBack;
  end;

var
  fWorkerForm: TfWorkerForm;

IMPLEMENTATION

{$R *.DFM}

function GenProgressCallBack(ecbType: CBType; Data: LongInt; pcbInfo: Pointer): CBRType; stdcall;
    var
        CBInfo: TUVerifyCallBack;
        
procedure UpdateProgress(progressbar: TProgressBar);
    begin
        if CBInfo.percentdone <> 65535 then
        begin
            if (CBInfo.percentdone < progressbar.position) then
            begin
                progressbar.Max := 100;
                progressbar.Position := 0;
            end;
            progressbar.position := CBInfo.percentdone; // round( CBInfo.percentdone - progressbar.position );
            progressbar.repaint;
        end
        else
        begin
            progressbar.position  := 100;
        end;
    end;

    begin
        CBInfo := TUVerifyCallBack(pcbInfo^);
//        dmfawp.FProcessedCount := 0;
        if ecbType = cbGENPROGRESS then
        case CBInfo.Process of
            TUVerifyHeader:
            begin
                UpdateProgress(fWorkerForm.pbVerifyHeader);
            end;
            TUVerifyIndex:
            begin
                UpdateProgress(fWorkerForm.pbVerifyIndex);
            end;
            TUVerifyData:
            begin
                UpdateProgress(fWorkerForm.pbVerifyData);
            end;
            TURebuild:
            begin
                UpdateProgress(fWorkerForm.pbRebuild);
            end;
        end;
        Application.ProcessMessages;
        Result := cbrUSEDEF;
    end;

constructor TBDEUtil.Create;
    begin
        Check(TUInit(vhtSes));
    end;

destructor TBDEUtil.Destroy;
    begin
        Check(TUExit(vhtSes));
        inherited Destroy;
    end;

function TBDEUtil.GetTCursorProps(szTable: String): Boolean;
    begin
        if TUFillCURProps(vHtSes, PChar(szTable), TUProps) = DBIERR_NONE then
          Result := True
        else Result := False;
    end;

procedure TBDEUtil.RegisterCallback;
    begin
        Check(DbiRegisterCallBack(nil, cbGENPROGRESS, 0,
        sizeof(TUVerifyCallBack), @CbInfo, GenProgressCallback));
    end;

procedure TBDEUtil.UnRegisterCallback;
    begin
        Check(DbiRegisterCallBack(nil, cbGENPROGRESS, 0,
        sizeof(TUVerifyCallBack), @CbInfo, nil));
    end;

procedure TfWorkerForm.AddLogMessage(S: string);
    begin
        mNotes.lines.add(s);
    end;

procedure TfWorkerForm.FormCreate(Sender: TObject);
    var
        i, j: integer;
        dsl: TDelimitedStringList;
    begin
        fClosing := False;
        CloseWhenDone := False;
        DefaultRepairType := rt_Verify;
        BDEUtil := TBDEUtil.Create;
        TablesList := TDelimitedStringList.Create;
        TablesList.HasFinalDelimiter := False;
        TablesList.Delimiters.Add(',');
        if ( (paramcount < 2) or (pos('?', paramstr(1)) <> 0) ) then
        begin
            MessageDlg('TableRepairCommand'+#13+#10+
            'usage: TableRepairCommand '+#13+#10+
            '              -T{ABLES} table1,table2,table3'+#13+#10+
            '              -V{ERIFY}|-R{EBUILD}|-A{UTO}'+#13+#10+
            '              -C{LOSE}'+#13+#10+
            '              -P{ASSWORDS} password1,password2,password3'+#13+#10+''+#13+#10+
            '-VERIFY means do a verify only'+#13+#10+
            '-REBUILD will do a verify, a forced rebuild, and then another verify if the rebuild succeeds'+#13+#10+
            '-AUTO will try to open the table, if that fails, it will do a forced rebuild. otherwise it will do a verify,'+#13+#10+
            '  and if that fails, a rebuild, and another verify'+#13+#10+
            '-CLOSE will exit automatically (with exitcode 1 if there was an error). '+#13+#10+
            'It will also create tablerepairerror.txt in the same directory as the exe'+#13+#10+
            'e.g'+#13+#10+
            'TableRepairCommand -t c:\tables\heads.db,c:\tables\tails.db VERFIY CLOSE -p secret',
            mtInformation, [mbOK], 0);
            PostQuitMessage( 1 );
        end
        else
        begin
            dsl := TDelimitedStringList.Create;
            try
                dsl.HasFinalDelimiter := false;
                dsl.Delimiters.Add(',');
                i := 1;
                while (i <= paramcount) do
                begin
                    if pos('-C', paramstr(i)) > 0 then
                      CloseWhenDone := True
                    else if pos('-V', uppercase(paramstr(i))) > 0 then
                      DefaultRepairType := rt_Verify
                    else if pos('-R', uppercase(paramstr(i))) > 0 then
                      DefaultRepairType := rt_Rebuild
                    else if pos('-A', uppercase(paramstr(i))) > 0 then
                      DefaultRepairType := rt_Auto
                    else if pos('-T', uppercase(paramstr(i))) > 0 then
                    begin
                        TablesList.DelimitedText := paramstr(i + 1);
                        inc(i);
                    end
                    else if pos('-P', uppercase(paramstr(i))) > 0 then
                    begin
                        DSL.DelimitedText := paramstr(i + 1);
                        for j := 0 to pred(DSL.Count) do
                          Session.AddPassword(DSL[j]);
                        inc(i);
                    end;
                    
                    inc(i);
                end;
            finally
                dsl.Clear;
                dsl.Free;
            end;
            PostMessage(Handle, WM_TIMER, tm_Init, 0);
        end;
    end;

procedure TfWorkerForm.FormDestroy(Sender: TObject);
    begin
        TablesList.Clear;
        TablesList.Free;
        BDEUtil.Free;
    end;

procedure TfWorkerForm.ResetProgressBar(which: integer);
    begin
        case which of
            1: pbVerifyHeader.Position := 0;
            2: pbVerifyIndex.Position := 0;
            3: pbVerifyData.Position := 0;
            4: pbRebuild.Position := 0;
        end;
    end;

function TfWorkerForm.VerifyTable(tablename: string;  var TableMessage: integer): boolean;
    var
        szTable: String;
        rslt: DBIResult;
    begin
        Result := False;
        Table1.DatabaseName := ExtractFilePath(Paramstr(0));
        Table1.TableName := 'test.db';
        if fileexists(table1.databasename + '\' + table1.tablename) then
          deletefile(table1.databasename + '\' + table1.tablename);
        Table1.CreateTable;
        Table1.Open;
        Table1.Close;
        Check(TUExit(BDEUtil.vHtSes));
        Check(TUInit(BDEUtil.vHtSes));
        szTable := tablename;
        BDEUtil.GetTCursorProps(szTable);
        BDEUtil.RegisterCallBack;
        try
            Check(TUVerifyTable(BDEUtil.vHtSes, PChar(szTable), szPARADOX, 'VERIFY.DB', nil, 0, TableMessage));
            rslt := TUGetCRTblDescCount(BDEUtil.vhTSes, PChar(szTable), iFld,
            iIdx, iSec, iVal, iRI, iOptP, iOptD);
            if rslt = DBIERR_NONE then
            begin
                case TableMessage of
                    0:
                    begin
                        if pbVerifyindex.Position = 0 then
                        begin
                            AddLogMessage(extractfilename(szTable) + ' Verification Successful, but table has no index.');
                        end
                        else if pbVerifyindex.Position < 100 then
                        begin
                            AddLogMessage(extractfilename(szTable) + ' Verification Successful, but index is incomplete.')
                        end
                        else
                        begin
                            AddLogMessage(extractfilename(szTable) + ' Verification Successful. Table has no errors.');
                            result := true;
                        end;
                    end;
                    1:
                    begin
                        if pbVerifyindex.Position = 0 then
                        begin
                            AddLogMessage(extractfilename(szTable) + ' Verification Completed, but table has no index.');
                        end
                        else if pbVerifyindex.Position < 100 then
                        begin
                            AddLogMessage(extractfilename(szTable) + ' Verification Completed, but index is incomplete.');
                        end
                        else
                        begin
                            AddLogMessage(extractfilename(szTable) + ' Verification Successful. Verification completed.');
                            result := true;
                        end;
                    end;
                    2: AddLogMessage(extractfilename(szTable) + ' Verification Successful. Verification could not be completed.');
                    3: AddLogMessage(extractfilename(szTable) + ' Verification Successful. Table must be rebuild manually.');
                    4: AddLogMessage(extractfilename(szTable) + ' Verification Successful. Table cannot be rebuilt.');
                    else
                      AddLogMessage(extractfilename(szTable) + ' Verification unsuccessful.');
                end;
            end;
        finally
            BDEUtil.UnRegisterCallBack;
        end;
        if fileexists(table1.databasename + '\' + table1.tablename) then
          deletefile(table1.databasename + '\' + table1.tablename);
    end;


function TfWorkerForm.RebuildTable(tablename: string; var TableMessage: integer): boolean;
    var
// use the global ones        iFld, iIdx, iSec, iVal, iRI, iOptP, iOptD: word;
        szTable: String;
        rslt: DBIResult;
        TblDesc: CRTBlDesc;
        Backup: String;
        szPassword : string;
    begin
        Result := False;
        Table1.DatabaseName := ExtractFilePath(Paramstr(0));
        Table1.TableName := 'test.db';
        if fileexists(table1.databasename + '\' + table1.tablename) then
          deletefile(table1.databasename + '\' + table1.tablename);
        Table1.CreateTable;
        Table1.Open;
        Table1.Close;
        Check(TUExit(BDEUtil.vHtSes));
        Check(TUInit(BDEUtil.vHtSes));
        szTable := tablename;
        BDEUtil.GetTCursorProps(szTable);
        BDEUtil.RegisterCallBack;
        try
            FillChar(TblDesc, SizeOf(CRTBlDesc), 0);
            StrPCopy(TblDesc.szTblName, szTable);
            TblDesc.szTblType := szParadox;
            TblDesc.szErrTblName := 'Rebuild.DB';
            TblDesc.iFldCount := iFld;
            GetMem(TblDesc.pFldDesc, (iFld * SizeOf(FldDesc)));
            
            TblDesc.iIdxCount := iIdx;
            GetMem(TblDesc.pIdxDesc, (iIdx * SizeOf(IdxDesc)));
            
            TblDesc.iSecRecCount := iSec;
            GetMem(TblDesc.pSecDesc, (iSec * SizeOf(SecDesc)));
            
            TblDesc.iValChkCount := iVal;
            GetMem(TblDesc.pvchkDesc, (iVal * SizeOf(VCHKDesc)));
            
            TblDesc.iRintCount := iRI;
            GetMem(TblDesc.printDesc, (iRI * SizeOf(RINTDesc)));
            
            TblDesc.iOptParams := iOptP;
            GetMem(TblDesc.pfldOptParams, (iOptP * sizeOf(FLDDesc)));
            
            GetMem(TblDesc.pOptData, (iOptD * DBIMAXSCFLDLEN));
            try
                rslt := TUFillCRTblDesc(BDEUtil.vhTSes, @TblDesc, PChar(szTable), pChar( szPassword ));
                if rslt = DBIERR_NONE then
                begin
                    Backup := 'Backup.Db';
                    if TURebuildTable(BDEUtil.vhTSes, PChar(szTable), szPARADOX,
                    PChar(Backup), 'KEYVIOL.DB', 'PROBLEM.DB', @TblDesc ) = DBIERR_NONE then
                    begin
                        AddLogMessage(extractfilename(szTable) + ' Rebuild succeeded');
                        Result := True;
                    end
                    else
                    begin
                        AddLogMessage(extractfilename(szTable) + ' Rebuild failed');
                    end;
                end;
            finally
                FreeMem(TblDesc.pFldDesc, (iFld * SizeOf(FldDesc)));
                FreeMem(TblDesc.pIdxDesc, (iIdx * SizeOf(IdxDesc)));
                FreeMem(TblDesc.pSecDesc, (iSec * SizeOf(SecDesc)));
                FreeMem(TblDesc.pvchkDesc, (iVal * SizeOf(VCHKDesc)));
                FreeMem(TblDesc.printDesc, (iRI * SizeOf(RINTDesc)));
                FreeMem(TblDesc.pfldOptParams, (iOptP * sizeOf(FLDDesc)));
                FreeMem(TblDesc.pOptData, (iOptD * DBIMAXSCFLDLEN));
            end;
        finally
            BDEUtil.UnRegisterCallBack;
        end;
        if fileexists(table1.databasename + '\' + table1.tablename) then
          deletefile(table1.databasename + '\' + table1.tablename);
    end;

procedure TfWorkerForm.WMTimer(var Msg: TMessage);
    var
        TableMessage: integer;
        i: integer;
        repairtype: integer; // 1 = verify, 2 = rebuild
    begin
        Application.ProcessMessages;
        if not fclosing then
        begin
            case Msg.wParam of
                tm_Init:
                begin
                    CurrentFileIndex := 0;
                    successcount := 0;
                    failedcount := 0;
                    PostMessage(Handle, WM_TIMER, tm_NextFile, 0);
                end;
                tm_NextFile:
                begin
                    if CurrentFileIndex >= TablesList.Count then
                    begin
                        AddLogMessage(
                        Format('---complete: %d files checked. %d succeeded, %d failed',
                        [TablesList.Count, successcount, failedcount]));
                        if  CloseWhenDone then
                        begin
                            if failedcount > 0 then
                            begin
                                mNotes.Lines.SaveToFile( ExtractFilePath(Paramstr(0)) + 'tablerepairerror.txt');
                                PostQuitMessage( 1 );
                            end
                            else
                              PostQuitMessage( 0 );
                        end;
                    end
                    else
                    begin
                        repairtype := DefaultRepairType;
                        if DefaultRepairType = rt_Auto then
                        begin
                            repairtype := 3; // auto repair if verify fails
                            try  // check each file for ".val/index is out of date" errors and force rebuild
                                tOpenAttempt.DatabaseName := ExtractFilePath(TablesList[CurrentFileIndex]);
                                tOpenAttempt.TableName := ExtractFileName(TablesList[CurrentFileIndex]);
                                tOpenAttempt.Open;
                            except
                                on e: exception do
                                begin
                                    AddLogMessage(extractfilename(TablesList[CurrentFileIndex]) + ' Rebuild required. ' + E.Message);
                                    repairtype := 2;
                                end;
                            end;
                            tOpenAttempt.Close;
                        end;

                        for i := 1 to 4 do ResetProgressBar(i); // reset all progress bars
                        case repairtype of
                            rt_Verify: // verify
                            begin
                                if VerifyTable(TablesList[CurrentFileIndex], TableMessage) then
                                  inc(successcount)
                                else
                                  inc(failedcount);
                            end;
                            rt_Rebuild: // force rebuild
                            begin
                                VerifyTable(TablesList[CurrentFileIndex], TableMessage);
                                begin
                                    for i := 3 to 4 do ResetProgressBar(i); // reset all progress bars
                                    if RebuildTable(TablesList[CurrentFileIndex], TableMessage) then
                                    begin
                                        for i := 1 to 3 do ResetProgressBar(i); // reset all progress bars
                                        if VerifyTable(TablesList[CurrentFileIndex], TableMessage) then
                                          inc(successcount)
                                        else
                                          inc(failedcount);
                                    end
                                    else
                                      inc(failedcount);
                                end;
                            end;
                            rt_Auto: // auto if verify fails
                            begin
                                if VerifyTable(TablesList[CurrentFileIndex], TableMessage) then
                                  inc(successcount)
                                else
                                begin
                                    for i := 3 to 4 do ResetProgressBar(i); // reset all progress bars
                                    if RebuildTable(TablesList[CurrentFileIndex], TableMessage) then
                                    begin
                                        for i := 1 to 3 do ResetProgressBar(i); // reset all progress bars
                                        if VerifyTable(TablesList[CurrentFileIndex], TableMessage) then
                                          inc(successcount)
                                        else
                                          inc(failedcount);
                                    end
                                    else
                                      inc(failedcount);
                                end;
                            end;
                            
                        end;
                        CurrentFileIndex := CurrentFileIndex + 1;
                        PostMessage(Handle, WM_TIMER, tm_NextFile, 0);
                    end;
                end;
                else
                  Raise Exception.Create('Invalid Timer message: ' + IntToStr( Msg.wParam ) );
            end;
        end;
    end;

procedure TfWorkerForm.FormClose(Sender: TObject;
    var Action: TCloseAction);
    begin
        Closing := True;
    end;

end.
